/* screenbuffer functions */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "config.h"
#include "console.h"
#include "cfg.h"


/* screen dimensions */
int SCREEN_X,SCREEN_Y;

int SCREEN_XOFFSET,SCREEN_YOFFSET;


/* screenbuffer */
unsigned char *screen;
unsigned char *screen_old;


/* attributes:
	lower 4 bits=foreground
	bits 4-7=background
*/
unsigned char *screen_a;
unsigned char *screen_a_old;


/* allocate memory for screenbuffer */
void init_blit(void)
{
	c_get_size(&SCREEN_X,&SCREEN_Y);
	screen=malloc(SCREEN_X*SCREEN_Y);
	if (!screen){ERROR("Not enough memory!\n");EXIT(1);}
	screen_a=malloc(SCREEN_X*SCREEN_Y);
	if (!screen_a){ERROR("Not enough memory!\n");EXIT(1);}
	screen_old=malloc(SCREEN_X*SCREEN_Y);
	if (!screen_old){ERROR("Not enough memory!\n");EXIT(1);}
	screen_a_old=malloc(SCREEN_X*SCREEN_Y);
	if (!screen_a_old){ERROR("Not enough memory!\n");EXIT(1);}
	memset(screen_a_old,0,SCREEN_X*SCREEN_Y);
	memset(screen_old,0,SCREEN_X*SCREEN_Y);
        SCREEN_XOFFSET=(SCREEN_X-PLAYER_WIDTH)>>1;
        SCREEN_YOFFSET=(SCREEN_Y-PLAYER_HEIGHT)>>1;
			
}


void clear_screen(void)
{
	memset(screen_a,0,SCREEN_X*SCREEN_Y);
	memset(screen,0,SCREEN_X*SCREEN_Y);
}


/* resize and clear screen buffers to new dimensions stored in SCREEN_X and SCREEN_Y*/
void resize_screen(void)
{
	c_get_size(&SCREEN_X,&SCREEN_Y);
	screen=realloc(screen,SCREEN_X*SCREEN_Y);
	if (!screen){ERROR("Not enough memory!\n");EXIT(1);}
	screen_a=realloc(screen_a,SCREEN_X*SCREEN_Y);
	if (!screen_a){ERROR("Not enough memory!\n");EXIT(1);}
	screen_old=realloc(screen_old,SCREEN_X*SCREEN_Y);
	if (!screen_old){ERROR("Not enough memory!\n");EXIT(1);}
	screen_a_old=realloc(screen_a_old,SCREEN_X*SCREEN_Y);
	if (!screen_a_old){ERROR("Not enough memory!\n");EXIT(1);}
	memset(screen_a_old,0,SCREEN_X*SCREEN_Y);
	memset(screen_old,0,SCREEN_X*SCREEN_Y);
	SCREEN_XOFFSET=(SCREEN_X-PLAYER_WIDTH)>>1;
	SCREEN_YOFFSET=(SCREEN_Y-PLAYER_HEIGHT)>>1;

	clear_screen();
}


void redraw_screen(void)
{
	memset(screen_a_old,0,SCREEN_X*SCREEN_Y);
	memset(screen_old,0,SCREEN_X*SCREEN_Y);
	clear_screen();
}


/* print text into screen buffer on given position, with given color */
void print2screen(int x,int y,unsigned char color,char *message)
{
	int offs=x+SCREEN_X*y;
	int l=strlen(message);

	color&=15;

	if (y>=SCREEN_Y||x>=SCREEN_X)return;
	if (l+offs>SCREEN_X*SCREEN_Y)l=SCREEN_X*SCREEN_Y-offs;
	
	memset(screen_a+offs,color,l);
	memcpy(screen+offs,message,l);
}


/* draw screenbuffer to the screen */
#ifdef HAVE_INLINE
	inline void 
#else
	void
#endif
blit_screen(void)
{
	int x,y;
	int yof;
	unsigned char a;
	int changed=1;
	int status_flag;
	unsigned char last_color=*screen_a;
	
	c_setcolor_bg(last_color,(*screen_a)>>4);
	for (y=0;y<SCREEN_Y;y++)
	{
		c_goto(0,y);
		yof=SCREEN_X*y;
		for (x=0;x<SCREEN_X;x++)
		{ 
			if (x == SCREEN_X - 1 && y == SCREEN_Y - 1) break;
			if (screen[x+yof]==screen_old[x+yof]&&screen_a[x+yof]==screen_a_old[x+yof]){changed=0;continue;}
			if (!changed)c_goto(x,y);
			changed=1;

			a=(last_color^(screen_a[x+yof]));
			status_flag=	(!!(a&7))|   /* foreground */
					(((a>>3)&1)<<1)|   /* foreground highlight */
					((!!((a>>4)&7))<<2);  /* background */

			switch(status_flag)  /* what changed: */
			{
				case 0:  /* nothing */
				break;

				case 1:  /* foreground */
				c_setcolor_3b(screen_a[x+yof]);
				break;

				case 6:  /* higlight and background */
				case 2:  /* highlight */
				if ((screen_a[x+yof])&8)  /* hlt on */
					c_sethlt_bg(((screen_a[x+yof])>>3)&1,(screen_a[x+yof])>>4);
				else
					c_setcolor_bg(screen_a[x+yof],(screen_a[x+yof])>>4);
				break;

				case 7: /* background, foreground and highlight */
				case 3: /* foreground and higlight */
				c_setcolor_bg(screen_a[x+yof],(screen_a[x+yof])>>4);
				break;
				
				case 4:  /* background */
				c_setbgcolor((screen_a[x+yof])>>4);
				break;

				case 5:  /* background and foreground */
				c_setcolor_3b_bg(screen_a[x+yof],(screen_a[x+yof])>>4);
				break;
			}
					
			last_color=screen_a[x+yof];
			c_putc(screen[x+SCREEN_X*y]?screen[x+SCREEN_X*y]:32);
		}
	}
	memcpy(screen_old,screen,SCREEN_X*SCREEN_Y);
	memcpy(screen_a_old,screen_a,SCREEN_X*SCREEN_Y);
	c_refresh();
}


/* draw frame to the screen */
/* x,y = upper left corner */
/* w,h = width and height of the content */
void draw_frame(int x,int y,int w,int h,unsigned char color)
{
	int a;
	char *t;
	
	color&=15;


	t=malloc(w+3);
	if (!t)return;
	t[w+2]=0;
	t[w]=0;
	t[w-1]=0;
	print2screen(x,y,color,"+");
	print2screen(x+w+1,y,color,"+");
	print2screen(x+w+1,y+h+1,color,"+");
	print2screen(x,y+h+1,color,"+");
	for (a=0;a<w;a++)
		t[a]='-';
	print2screen(x+1,y,color,t);
	print2screen(x+1,y+h+1,color,t);
	memset(t,' ',w+1);
	t[0]='|';
	t[w+1]='|';
	for (a=1;a<=h;a++)
		print2screen(x,y+a,color,t);
	free(t);
}
