#include <stdio.h>
#include <stdlib.h>

#include "sprite.h"
#include "console.h"
#include "blit.h"
#include "data.h"
#include "cfg.h"
#include "hash.h"
#include "time.h"


unsigned long id;   /* id of last object */
int xoffset=0,yoffset=0; /* coordinates of upper left corner of the screen */
int x=0,y=0;  /* cursor coordinates on the screen */
int oldx=0,oldy=0;  /* old cursor position */

struct object_list objects;
struct object_list *last_obj;


/* load static data */
void load_room(char * filename)
{
	FILE * stream;
	static char line[1024];
	char *p,*q,*name;
	int n,x,y,t;

	if (!(stream=fopen(filename,"rb")))
	{
		char msg[256];
		sprintf(msg,"Can't open file \"%s\"!\n",filename);
		ERROR(msg);
		EXIT(1);
	}
	while(fgets(line,1024,stream))
	{
		p=line;
		_skip_ws(&p);
		for (name=p;(*p)!=' '&&(*p)!=9&&(*p)!=10&&(*p);p++);
		if (!(*p))continue;
		*p=0;p++;
		_skip_ws(&p);
		t=*p;
		p++;
		_skip_ws(&p);
		x=strtol(p,&q,0);
		_skip_ws(&q);
		y=strtol(q,&p,0);
		if (find_sprite(name,&n))
		{
			char msg[256];
			sprintf(msg,"Unknown bitmap name \"%s\"!\n",name);
			ERROR(msg);
			EXIT(1);
		}
		new_obj(id,t,0,n,0,0,x,y,0,0,0);
		id++;
	}
	fclose(stream);
}



/* draw view */
void draw_scene(void)
{
	struct object_list *p;
	
	for(p=&objects;p->next;p=p->next)
	{
		put_sprite(
			p->next->member.x-xoffset,
			p->next->member.y-yoffset,
			sprites[p->next->member.sprite].positions,0
			);
	}
}


/* find object under cursor */
struct it * find_object(void)
{
	struct object_list *p;
#define O p->next->member
	
	for(p=&objects;p->next;p=p->next)
	{
		if (
			O.x<=x+xoffset&&
			O.x+sprites[O.sprite].positions[0].lines[0].len-1>=x+xoffset&&
			O.y<=y+yoffset&&
			O.y+sprites[O.sprite].positions[0].n-1>=y+yoffset
			)
			return &(p->next->member);
	}
	return 0;
}


/* next type when 't' is pressed */
int next_type(int type)
{
	switch(type)
	{
		case 'w':
		return 'b';

		case 'b':
		return 'f';

		case 'f':
		return 'j';

		case 'j':
		return 'i';

		case 'i':
		return 'w';

		default:
		return 'b';
	}
}


/* save level */
void save_data(char *name)
{
	FILE *stream;
	struct object_list *p;
	static unsigned char txt[1024];

	stream=fopen(name,"w");
	if (!stream)
	{
		char msg[256];
		sprintf(msg,"Can't create file \"%s\"\n",name);
		ERROR(msg);
		EXIT(1);
	}
	
#define O p->next->member
	
	for(p=&objects;p->next;p=p->next)
	{
		sprintf(txt,"%s %c %d %d\n",sprite_names[O.sprite],O.type,O.x,O.y);
		fputs(txt,stream);
	}
	fclose(stream);
}


int main(int argc, char**argv)
{
	int update=1;
	unsigned char bg=' ';
	unsigned char color=0;
	struct it* obj=0;
	char txt[256];
	unsigned long_long last_time;

	last_obj=&objects;

	if (argc>2)
	{
		SCREEN_X=strtoul(argv[1],0,10);
		SCREEN_Y=strtoul(argv[2],0,10);
	}
	hash_table_init();
	init_sprites();
	init_area();
	load_sprites(SPRITES_FILE);
	load_room(DATA_FILE);
	c_init(SCREEN_X,SCREEN_Y);
	c_cursor(C_HIDE);
	update=1;
again:
	last_time=get_time();

	switch(update)
	{
		case 1:
		clear_screen();
		draw_scene();
		blit_screen();
		
		case 0:
		if (!update)
		{
			c_goto(oldx,oldy);
			c_setcolor(color);
			c_putc(bg?bg:' ');
		}
		color=screen_a[x+y*SCREEN_X];
		bg=screen[x+y*SCREEN_X];

		c_goto(x,y);
		c_setcolor(15);
		c_print("+");
		break;

	}
	obj=find_object();
	c_clear(0,SCREEN_Y-1,SCREEN_X-2,SCREEN_Y-1);
	c_goto(0,SCREEN_Y-1);
	c_setcolor(11);
	c_print("X: ");
	sprintf(txt,"% 4d   ",x+xoffset);
	c_setcolor(7);
	c_print(txt);
	c_setcolor(11);
	c_print("Y: ");
	sprintf(txt,"% 4d   ",y+yoffset);
	c_setcolor(7);
	c_print(txt);
	c_setcolor(11);
	c_print("OBJECT: ");
	sprintf(txt,"%-20s   ",obj?sprite_names[obj->sprite]:"----");
	c_setcolor(7);
	c_print(txt);
	c_setcolor(11);
	c_print("TYPE: ");
	sprintf(txt,"%c   ",obj?obj->type:'-');
	c_setcolor(7);
	c_print(txt);
	update=0;
	oldx=x;
	oldy=y;
	c_refresh();
	
	c_update_kbd();

	if (c_pressed('o'))
	{
		xoffset-=SCREEN_X>>2;
		if (xoffset<0)xoffset=0;
		else update=1;
	}
		
	if (c_pressed('p'))
	{
		xoffset+=SCREEN_X>>2;
		if (xoffset>AREA_X-1-(SCREEN_X>>1))xoffset=AREA_X-1-(SCREEN_X>>1);
		update=1;
	}
		
	if (c_pressed('a'))
	{
		yoffset-=SCREEN_Y>>2;
		if (yoffset<0)yoffset=0;
		update=1;
	}
		
	if (c_pressed('z'))
	{
		yoffset+=SCREEN_Y>>2;
		if (yoffset>AREA_Y-1-(SCREEN_Y>>1))yoffset=AREA_Y-1-(SCREEN_Y>>1);
		update=1;
	}
		
	if (c_pressed('q')||c_pressed(K_ESCAPE))
	{
		c_shutdown();
		exit(0);
	}
	
	if (c_pressed(K_RIGHT))
	{
		x++;
		if (x>SCREEN_X-1||x+xoffset>AREA_X-1)x=0;
	}

	if (c_pressed(K_LEFT))
	{
		x--;
		if (x<0)x=SCREEN_X-1;
		if (x+xoffset>AREA_X-1)x=AREA_X-1-xoffset;
	}

	if (c_pressed(K_DOWN))
	{
		y++;
		if (y>SCREEN_Y-1||y+yoffset>AREA_Y-1)y=0;
	}

	if (c_pressed(K_UP))
	{
		y--;
		if (y<0)y=SCREEN_Y-1;
		if (y+yoffset>AREA_Y-1)y=AREA_Y-1-yoffset;
	}

	if (c_was_pressed('x'))
		if (obj)
		{
			delete_obj(obj->id);
			update=1;
		}

	if (c_was_pressed('=')||c_was_pressed(K_NUM_PLUS)|c_was_pressed(K_PGUP))
		if (obj)
		{
			obj->sprite++;
			obj->sprite%=n_sprites;
			update=1;
		}
	
	if (c_was_pressed('l')||((c_pressed(K_LEFT_SHIFT)||c_pressed(K_RIGHT_SHIFT))&&c_pressed('l')))
		if (obj)
		{
			obj->x--;
			if (obj->x<0)obj->x=0;
			update=1;
		}

	if (c_was_pressed('r')||((c_pressed(K_LEFT_SHIFT)||c_pressed(K_RIGHT_SHIFT))&&c_pressed('r')))
		if (obj)
		{
			obj->x++;
			if (obj->x>AREA_X-1)obj->x=AREA_X-1;
			update=1;
		}

	if (c_was_pressed('u')||((c_pressed(K_LEFT_SHIFT)||c_pressed(K_RIGHT_SHIFT))&&c_pressed('u')))
		if (obj)
		{
			obj->y--;
			if (obj->y<0)obj->y=0;
			update=1;
		}

	if (c_was_pressed('d')||((c_pressed(K_LEFT_SHIFT)||c_pressed(K_RIGHT_SHIFT))&&c_pressed('d')))
		if (obj)
		{
			obj->y++;
			if (obj->y>AREA_Y-1)obj->y=AREA_Y-1;
			update=1;
		}
	
	if (c_was_pressed('-')||c_was_pressed(K_NUM_MINUS)||c_was_pressed(K_PGDOWN))
		if (obj)
		{
			if (!obj->sprite)obj->sprite=n_sprites-1;
			else obj->sprite--;
			update=1;
		}

	if (c_was_pressed('t')||c_was_pressed(K_TAB))
		if (obj)
		{
			obj->type=next_type(obj->type);
			update=1;
		}

	if (c_was_pressed('n')||c_was_pressed(K_ENTER))
	{
		new_obj(id,'w',0,0,0,0,x+xoffset,y+yoffset,0,0,0);
		id++;
		update=1;
	}

	if (c_was_pressed('s')||c_was_pressed(K_F2))
		save_data(DATA_FILE);

	sleep_until(last_time+1000000/35);
	goto again;
}
