#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#ifndef WIN32
#include "config.h"
#endif

#include "sprite.h"
#include "blit.h"
#include "data.h"
#include "console.h"
#include "cfg.h"



/* allocate memory for screenbuffer */
void init_sprites(void)
{
	init_blit();
}


/* write visible window of playing area into screenbuffer */
void show_window(int x,int y)
{
	int a,loffs,roffs,b;
 
	if (x+SCREEN_X<=0)return;
	if (y+SCREEN_Y<=0)return;

	loffs=x<0?-x:0;
	roffs=x+SCREEN_X>AREA_X?x+SCREEN_X-AREA_X:0;
	for(a=(y<0?-y:0);a<SCREEN_Y&&(y+a)<AREA_Y;a++)
	{
		memcpy(screen+a*SCREEN_X+loffs,area+loffs+x+(y+a)*AREA_X,SCREEN_X-loffs-roffs);
		for(b=0;b<SCREEN_X-loffs-roffs;b++)
			screen_a[a*SCREEN_X+loffs+b]=((area_a[loffs+x+(y+a)*AREA_X+b])&15);
	}
}


/* write one scanline of a bitmap to given memory */
#ifdef HAVE_INLINE
	inline void 
#else
	void
#endif
put_line(int xs, int ys,char *scrn,char *attr,int x,int y,struct line *l,int type,char plastic)
{
	int a;
	if (y>=ys||y<0)return;
	for (a=0;a<l->len&&x+a<xs;a++)
	{
		if (!l->attr[a])continue;
		if (plastic&&(((attr[x+a+xs*y]&240)==TYPE_FOREGROUND)||((attr[x+a+xs*y]&240)==TYPE_JUMP_FOREGROUND)))continue;
		if (x+a<0)continue;
		scrn[x+a+xs*y]=l->bmp[a];
		attr[x+a+xs*y]=(l->attr[a])|type;
	}
}

/* put image to given memory */
#ifdef HAVE_INLINE
	inline void 
#else
	void
#endif
_put_sprite(int xs, int ys, char *scrn,char *attr,int x,int y,struct pos *p,int type,char plastic)
{
	int a;
	for (a=0;a<p->n;a++)
		put_line(xs,ys,scrn,attr,x+p->xo,a+y+p->yo,p->lines+a,type,plastic);
}

/* put image into screenbuffer */
#ifdef HAVE_INLINE
	inline void
#else
	void 
#endif
put_sprite(int x, int y, struct pos *p,char plastic)
{
	_put_sprite(SCREEN_X,SCREEN_Y,screen,screen_a,x,y,p,0,plastic);
}


/* ancillary function for converting color from file notation to value 0-15 */
int _conv_color(int c)
{
	int a=tolower(c);
	if (a>='a'&&a<='f')return 10+a-'a';
	if (a>='0'&&a<='9')return a-'0';
	/* space or other character is blank attr */
	return 0;
}


/* load sprite from a file */
void load_sprite(char * filename,struct sprite *s)
{
#define CURP (s->n_positions-1)
#define CURL (s->positions[CURP].n-1)
	FILE *f;
	int x,a;
	static char buffer[8192];
	char *p,*q;
	int step=0;  /*0=expecting 'p', 1=expecting 'l', 2=expecting 'a', 3=expecting 'p' or 's' or 'l', 4=expecting end */

	s->positions=0;
	s->n_positions=0;
 
	if (!(f=fopen(filename,"rb")))
	{
		char msg[256];
		sprintf(msg,"Error opening file \"%s\"!\n",filename);
		ERROR(msg);
		EXIT(1);
	}
	while(fgets(buffer,8191,f))
	{
		x=strlen(buffer);
		if (buffer[x-1]==10)buffer[x-1]=0;
		switch(*buffer)
		{
			case 0:
			goto skip;

			case '#': /* comment */
			break;

			case 'p':
			if (step!=0&&step!=3)
			{
				char msg[256];
				sprintf(msg,"Syntax error in file \"%s\".\n",filename);
				ERROR(msg);
				EXIT(1);
			}
			step=1;
			s->n_positions++;
			s->positions=realloc(s->positions,s->n_positions*sizeof(struct pos));
			if (!s->positions)
			{
				ERROR("Memory allocation error!\n");
				EXIT(1);
			}
			s->positions[CURP].n=0;
			s->positions[CURP].lines=0;
			s->n_steps=0;
			s->steps=0;
			s->positions[CURP].xo=strtol(buffer+1,&p,0);
			s->positions[CURP].yo=strtol(p+1,&q,0);
			break;
   
			case 'l':
			if (step!=1&&step!=3)
			{
				char msg[256];
				sprintf(msg,"Syntax error in file \"%s\".\n",filename);
				ERROR(msg);
				EXIT(1);
			}
			step=2;
			s->positions[CURP].n++;
			s->positions[CURP].lines=realloc(s->positions[CURP].lines,s->positions[CURP].n*sizeof(struct line));
			if (!s->positions[CURP].lines)
			{
				ERROR("Memory allocation error!\n");
				EXIT(1);
			}
			s->positions[CURP].lines[CURL].len=strlen(buffer+1);
			if (!s->positions[CURP].lines[CURL].len)break;  /* empty line, we have nothing to do */
			s->positions[CURP].lines[CURL].bmp=malloc(s->positions[CURP].lines[CURL].len);
			if (!s->positions[CURP].lines[CURL].bmp)
			{
				ERROR("Memory allocation error!\n");
				EXIT(1);
			}
			memcpy(s->positions[CURP].lines[CURL].bmp,buffer+1,s->positions[CURP].lines[CURL].len);
			break;
   
			case 'a':
			if (step!=2)
			{
				char msg[256];
				sprintf(msg,"Syntax error in file \"%s\".\n",filename);
				ERROR(msg);
				EXIT(1);
			}
			step=3;
			/* UGH, co to tady je??????? */
			if (!s->positions[CURP].lines)
			{
				ERROR("Memory allocation error!\n");
				EXIT(1);
			}
			if (!s->positions[CURP].lines[CURL].len)break;  /* empty line, we have nothing to do */
			s->positions[CURP].lines[CURL].attr=malloc(s->positions[CURP].lines[CURL].len);
			if (!s->positions[CURP].lines[CURL].attr)
			{
				ERROR("Memory allocation error!\n");
				EXIT(1);
			}
			x=strlen(buffer+1);
			if (x>s->positions[CURP].lines[CURL].len)x=s->positions[CURP].lines[CURL].len;
			for (a=0;a<x;a++)
				s->positions[CURP].lines[CURL].attr[a]=_conv_color(buffer[1+a]);
			memset(s->positions[CURP].lines[CURL].attr+x,0,s->positions[CURP].lines[CURL].len-x);
			break;
   
			case 's':
			if (step!=3)
			{
				char msg[256];
				sprintf(msg,"Syntax error in file \"%s\".\n",filename);
				ERROR(msg);
				EXIT(1);
			}
			step=4;
			for(q=buffer,p=buffer+1;*q&&*q!=10;*q==','?p=q+1:0)
			{
				x=strtol(p,&q,0);
				s->n_steps++;
				s->steps=realloc(s->steps,s->n_steps*sizeof(unsigned short));
				if (!s->steps)
				{
					ERROR("Memory allocation error!\n");
					EXIT(1);
				}
				s->steps[s->n_steps-1]=x;
			}
			break;

			default:
			{
				char msg[256];
				sprintf(msg,"Syntax error in file \"%s\"!\n",filename);
				ERROR(msg);
				EXIT(1);
			}
		}
skip:  ;
	}
	if (step!=4)
	{
		char msg[256];
		sprintf(msg,"Unexpected end of file in \"%s\".\n",filename);
		ERROR(msg);
		EXIT(1);
	}
	fclose(f);
}
